@@ -10,7 +10,6 @@ import btrfs
 import Butter
 import Store
 
-import datetime
 import logging
 import math
 import os
@@ -22,44 +21,6 @@ logger = logging.getLogger(__name__)
 theMinimumChangeRate = .00001
 
 
-class _Writer:
-
-    """ Context Manager to write a snapshot. """
-
-    def __init__(self, process, path):
-        self.process = process
-        self.stream = process.stdin
-        self.path = path
-
-    def __enter__(self):
-        return self.stream.__enter__()
-
-    def __exit__(self, exceptionType, exception, trace):
-        self.stream.__exit__(exceptionType, exception, trace)
-
-        logger.debug("Waiting for receive process...")
-        self.process.wait()
-
-        if exception is None and self.process.returncode == 0:
-            return
-
-        if os.path.exists(self.path):
-            # This tries to mark partial (failed) transfers.
-
-            partial = self.path + ".part"
-
-            if os.path.exists(partial):
-                partial = self.path + "_" + datetime.datetime.now().isoformat() + ".part"
-
-            os.rename(self.path, partial)
-
-        if exception is None:
-            raise Exception(
-                "receive returned error %d. %s may be corrupt."
-                % (self.process.returncode, self.path)
-                )
-
-
 class ButterStore(Store.Store):
 
     """ A local btrfs synchronization source or sink. """
@@ -85,6 +46,8 @@ class ButterStore(Store.Store):
         self.btrfs = btrfs.FileSystem(self.userPath)     # ioctl interface
 
         self.butterVolumes = {}   # Dict of {uuid: <btrfs.Volume>}
+        self.extraVolumes = {}
+
         self._fillVolumesAndPaths()
 
     def _btrfsVol2StoreVol(self, bvol):
@@ -110,6 +73,8 @@ class ButterStore(Store.Store):
                     logger.warn("Skipping deleted volume %s", bv.uuid)
                     continue
 
+                relPath = None
+
                 # vol = Store.Volume(uuid, bv.totalSize, bv.exclusiveSize, bv.gen)
                 for path in bv.linuxPaths:
                     path = self._relativePath(path)
@@ -117,12 +82,22 @@ class ButterStore(Store.Store):
                     if path is None:
                         continue
 
-                    logger.debug("%s %s", vol, path)
                     self.paths[vol].append(path)
 
+                    infoPath = self._fullPath(path + ".bs")
+                    if os.path.exists(infoPath):
+                        logger.debug("Reading %s", infoPath)
+                        with open(infoPath) as info:
+                            Store.Volume.readInfo(info)
+
+                    if not path.startswith("/"):
+                        relPath = path
+
                 if vol not in self.paths:
                     continue
 
+                logger.debug("%s", vol.display(sink=self, detail='phrase'))
+
                 if vol.uuid in self.butterVolumes:
                     logger.warn(
                         "Duplicate effective uuid %s in '%s' and '%s'",
@@ -131,6 +106,9 @@ class ButterStore(Store.Store):
 
                 self.butterVolumes[vol.uuid] = bv
 
+                if relPath is not None:
+                    self.extraVolumes[vol] = relPath
+
     def _fileSystemSync(self):
         with self.btrfs as mount:
             mount.SYNC()
@@ -175,7 +153,7 @@ class ButterStore(Store.Store):
 
             toVol = self._btrfsVol2StoreVol(toBVol)
 
-            yield Store.Diff(self, toVol, fromVol, estimatedSize, True)
+            yield Store.Diff(self, toVol, fromVol, estimatedSize, sizeIsEstimated=True)
 
     def hasEdge(self, diff):
         """ True if Store already contains this edge. """
@@ -188,9 +166,7 @@ class ButterStore(Store.Store):
 
         path = self.selectReceivePath(paths)
 
-        process = self.butter.processReceive(os.path.dirname(path))
-
-        return _Writer(process, path) if process is not None else None
+        return self.butter.receive(path)
 
     def receiveVolumeInfo(self, paths):
         """ Return Context Manager for a file-like (stream) object to store volume info. """
@@ -210,7 +186,7 @@ class ButterStore(Store.Store):
         estimatedSize += toBVol.totalSize * (1 - math.exp(-changeRate * genDiff))
         estimatedSize = max(toBVol.exclusiveSize, estimatedSize)
 
-        return estimatedSize
+        return 2 * estimatedSize
 
     def _calcChangeRate(self, bvols):
         total = 0
@@ -235,21 +211,56 @@ class ButterStore(Store.Store):
             rate = - math.log(1 - diffs / total) * (len(bvols) - 1) / (maxGen - minGen)
             rate /= 10  # Fudge
         except (ZeroDivisionError, ValueError):
-            logger.debug("Using minimum change rate.")
+            # logger.debug("Using minimum change rate.")
             rate = theMinimumChangeRate
 
-        logger.debug("Change rate: %f", rate)
+        # logger.debug("Change rate: %f", rate)
 
         return rate
 
-    def send(self, diff, streamContext, progress=True):
+    def send(self, diff, progress=True):
         """ Write the diff (toVol from fromVol) to the stream context manager. """
         if not self.dryrun:
             self._fileSystemSync()
 
-        self.butter.send(
+        return self.butter.send(
             self.getSendPath(diff.toVol),
             self.getSendPath(diff.fromVol),
-            streamContext,
-            progress,
         )
+
+    def keep(self, diff):
+        """ Mark this diff (or volume) to be kept in path. """
+        self._keepVol(diff.toVol)
+        self._keepVol(diff.fromVol)
+
+    def _keepVol(self, vol):
+        """ Mark this volume to be kept in path. """
+        if vol is None:
+            return
+
+        if vol in self.extraVolumes:
+            del self.extraVolumes[vol]
+            return
+
+        newPath = self.selectReceivePath(self.paths[vol])
+
+        if self._skipDryRun(logger)("Copy %s to %s", vol, newPath):
+            return
+
+        self.butterVolumes[vol.uuid].copy(newPath)
+
+    def deleteUnused(self):
+        """ Delete any old snapshots in path, if not kept. """
+        for (vol, path) in self.extraVolumes.items():
+            if self._skipDryRun(logger)("Delete subvolume %s", path):
+                continue
+            self.butterVolumes[vol.uuid].destroy()
+
+    def deletePartials(self):
+        """ Delete any old partial uploads/downloads in path. """
+        for (vol, path) in self.extraVolumes.items():
+            if not path.endswith(".part"):
+                continue
+            if self._skipDryRun(logger)("Delete subvolume %s", path):
+                continue
+            self.butterVolumes[vol.uuid].destroy()
