@@ -10,6 +10,7 @@ if True:  # Headers
     if True:  # imports
 
         import argparse
+        import errno
         import logging
         import os.path
         import re
@@ -33,15 +34,16 @@ except IOError:
     print("Missing '%s'" % (theVersionFile))
     theVersion = "<unknown>"
 
+theChunkSize = 100
+
 command = argparse.ArgumentParser(
     description="Synchronize two sets of btrfs snapshots.",
     epilog="""
-<src>, <dst>:   [file://]/path/to/directory/[snapshot]
-                ssh://[user@]host/path/to/directory (Not implemented)
+<src>, <dst>:   [btrfs://]/path/to/directory/[snapshot]
                 s3://bucket/prefix/[snapshot]
 
-If only <dst> is supplied, just list available snapshots.  The trailing "/"
-on a <src> *is* significant.
+If only <dst> is supplied, just list available snapshots.  NOTE: The trailing
+"/" *is* significant.
 
 Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
 See README.md and LICENSE.txt for more info.
@@ -58,7 +60,7 @@ command.add_argument('-n', '--dry-run', action="store_true",
                      help="display what would be transferred, but don't do it",
                      )
 command.add_argument('-d', '--delete', action="store_true",
-                     help='delete any snapshots in <dst> that are not in <src> (Not implemented)',
+                     help='delete any snapshots in <dst> that are not in <src>',
                      )
 
 command.add_argument('-q', '--quiet', action="count", default=0,
@@ -66,13 +68,17 @@ command.add_argument('-q', '--quiet', action="count", default=0,
                      once: don't display progress.
                      twice: only display error messages""",
                      )
-command.add_argument('-l', '--logfile', type=argparse.FileType('w'),
+command.add_argument('-l', '--logfile', type=argparse.FileType('a'),
                      help='log debugging information to file',
                      )
 command.add_argument('-V', '--version', action="version", version='%(prog)s ' + theVersion,
                      help='display version',
                      )
 
+command.add_argument('--part-size', action="store", type=int, default=theChunkSize,
+                     help='Size of chunks in a multipart upload',
+                     )
+
 command.add_argument('--remote-receive', action="store_true",
                      help=argparse.SUPPRESS,
                      )
@@ -99,6 +105,7 @@ def _setupLogging(quietLevel, logFile):
     root.setLevel("INFO" if theDebug else "DEBUG")  # When debugging, this is handled per-logger
 
     def add(handler, level, format):
+        handler = logging.StreamHandler(handler)
         handler.setLevel(level)
         handler.setFormatter(logging.Formatter(format))
         root.addHandler(handler)
@@ -106,10 +113,12 @@ def _setupLogging(quietLevel, logFile):
     level = "DEBUG" if theDebug else "INFO" if quietLevel < 2 else "WARN"
     formatString = theDebugDisplayFormat if theDebug else theDisplayFormat
 
-    add(logging.StreamHandler(sys.stdout), level, formatString)
+    add(sys.stdout, level, formatString)
 
     if logFile is not None:
-        add(logging.StreamHandler(logFile), "DEBUG", theLogFormat)
+        add(logFile, "DEBUG", theLogFormat)
+
+    logging.getLogger('boto').setLevel("WARN")
 
 
 def parseSink(uri, isDest, dryrun):
@@ -126,15 +135,16 @@ def parseSink(uri, isDest, dryrun):
     parts = match.groupdict()
 
     if parts['method'] is None:
-        parts['method'] = 'file'
+        parts['method'] = 'btrfs'
 
-    if parts['method'] == 'file':
+    if parts['method'] in ('btrfs', 'file'):
         parts['path'] = parts['host'] + '/' + parts['path']
 
     logger.debug(parts)
 
     Sinks = {
-        'file': ButterStore.ButterStore,
+        'btrfs': ButterStore.ButterStore,
+        # 'file': FileStore,
         's3': S3Store.S3Store,
         # 'ssh': SSHStore.SSHStore,
     }
@@ -144,64 +154,73 @@ def parseSink(uri, isDest, dryrun):
 
 def main():
     """ Main program. """
-    args = command.parse_args()
+    try:
+        args = command.parse_args()
 
-    _setupLogging(args.quiet, args.logfile)
+        _setupLogging(args.quiet, args.logfile)
 
-    logger.debug("Arguments: %s", vars(args))
+        logger.debug("Arguments: %s", vars(args))
 
-    progress = args.quiet == 0
+        progress = args.quiet == 0
 
-    source = parseSink(args.source, False, args.dry_run)
+        source = parseSink(args.source, False, args.dry_run)
 
-    dest = parseSink(args.dest, source is not None, args.dry_run)
+        dest = parseSink(args.dest, source is not None, args.dry_run)
+
+        if source is None:
+            source = dest
+            dest = None
+
+        try:
+            next(source.listVolumes())
+        except StopIteration:
+            logger.error("No snapshots in source.")
+            path = args.source or args.dest
+            if not path.endswith("/"):
+                logger.error("Try adding a '/' to '%s'.", path)
+            return 1
+
+        if dest is None:
+            for item in source.listContents():
+                print item
+            if args.delete:
+                source.deletePartials()
+            return 0
+
+        best = BestDiffs.BestDiffs(source.listVolumes(), args.delete)
+        best.analyze(source, dest)
+
+        summary = best.summary()
+        logger.info("Optimal synchronization:")
+        for sink, values in summary.items():
+            logger.info("%s from %d diffs in %s",
+                        Store.humanize(values.size),
+                        values.count,
+                        sink or "TOTAL",
+                        )
+
+        for diff in best.iterDiffs():
+            if diff is None:
+                raise Exception("Missing diff.  Can't fully replicate.")
+            else:
+                diff.sendTo(dest, chunkSize=args.part_size << 20, progress=progress)
+
+        if args.delete:
+            dest.deleteUnused()
+
+        logger.debug("Successful exit")
 
-    if source is None:
-        for vol in dest.listVolumes():
-            print vol.display(dest, detail="line")
         return 0
-
-    vols = source.listVolumes()
-
-    best = BestDiffs.BestDiffs(vols, args.delete)
-    best.analyze(source, dest)
-
-    summary = best.summary()
-    logger.info("Optimal synchronization:")
-    for sink, values in summary.items():
-        logger.info("%s from %d diffs in %s",
-                    Store.humanize(values.size),
-                    values.count,
-                    sink or "TOTAL",
-                    )
-
-    for diff in best.iterDiffs():
-        if diff is None:
-            raise Exception("Missing diff.  Can't fully replicate.")
-            continue
-
-        logger.info("%s: %s", "Keep" if diff.sink == dest else "Xfer", diff)
-
-        vol = diff.toVol
-        paths = diff.sink.getPaths(vol)
-
-        if diff.sink != dest:
-            streamContext = dest.receive(diff, paths)
-
-            diff.sink.send(diff, streamContext, progress=progress)
-
-            # TODO: For symmetry, put this into the streamContext.__exit__ method
-
-            if vol.hasInfo():
-                infoContext = dest.receiveVolumeInfo(paths)
-
-                if args.dry_run:
-                    vol.writeInfo(sys.stdout)
-                else:
-                    with infoContext as stream:
-                        vol.writeInfo(stream)
-
-    return 0
+    except Exception as error:
+        if (
+            isinstance(error, IOError) and
+            error.errno == errno.EPERM and
+            os.getuid() != 0
+        ):
+            logger.error("You must be root to access a btrfs filesystem.  Use  'sudo'")
+        else:
+            logger.exception("")
+        return 1
 
 if __name__ == "__main__":
     sys.exit(main())
